
/* matmatmult.c */

#include <stdio.h>
#include <stdlib.h>
#include "second.h"

const int minimum_matrix_size     = 1;
const int minimum_multiply_method = 1;
const int by_init      = minimum_multiply_method;
const int by_tiling    = by_init + 1;
const int maximum_multiply_method = by_tiling;
const int program_failure_code = -1;
const int program_success_code =  0;

static int product_rows, product_columns, product_betweens;
static int multiply_method;
static int rows_per_tile, columns_per_tile, betweens_per_tile;

static void input_sizes(void);
static float** matrix_allocate(int matrix_rows, int matrix_columns);
static float** matrix_deallocate(float** matrix,
                                 int matrix_rows, int matrix_columns);
static void matrix_set_randomly(float** matrix,
                                int number_of_rows, int number_of_columns);
static void matrix_matrix_mult_by_init(
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens);
static void matrix_matrix_mult_by_tiling(
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens,
         int rows_per_tile, int columns_per_tile, int betweens_per_tile);
static void matrix_matrix_multiply_tile(
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens,
         int row_start,     int row_end,
         int column_start,  int column_end,
         int between_start, int between_end);

int main ()
{ /* main */

    float** product = (float**)NULL;
    float** matrix1 = (float**)NULL;
    float** matrix2 = (float**)NULL;

   /*
    * Input the sizes of the matrices, the type of multiplication,
    * and the sizes of the tiles (if tiling).
    */

    input_sizes();

   /*
    * Allocate the matrices.
    */

    product = matrix_allocate(product_rows,     product_columns);

    matrix1 = matrix_allocate(product_rows,     product_betweens);

    matrix2 = matrix_allocate(product_betweens, product_columns);

   /*
    * Set the multiplier and multiplicand matrices at random.
    */

    matrix_set_randomly(matrix1, product_rows,     product_betweens);

    matrix_set_randomly(matrix2, product_betweens, product_columns);

   /*
    * Multiply the matrices by the appropriate method.
    */

    if (multiply_method == by_init) {

        matrix_matrix_mult_by_init(
            product, matrix1, matrix2,
            product_rows, product_columns, product_betweens);

    } /* if (multiply_method == by_init) */
    else if (multiply_method == by_tiling) {

        matrix_matrix_mult_by_tiling(
            product, matrix1, matrix2,
            product_rows, product_columns, product_betweens,
            rows_per_tile, columns_per_tile, betweens_per_tile);

    } /* if (multiply_method == by_tiling) */

   /*
    * Deallocate the matrices.
    */

    product =
        matrix_deallocate(product, product_rows,     product_columns);
    matrix1 =
        matrix_deallocate(matrix1, product_rows,     product_betweens);
    matrix2 =
        matrix_deallocate(matrix2, product_betweens, product_columns);

} /* main */

void input_sizes (void)
{ /* input_sizes */

    scanf("%d %d %d",
          &product_rows, &product_columns, &product_betweens);

    if (product_rows < minimum_matrix_size) {

        fprintf(stderr, "ERROR: cannot have product rows of %d.\n",
            product_rows);

        exit(program_failure_code);

    } /* if (product_rows < minimum_matrix_size) */

    if (product_columns < minimum_matrix_size) {

        fprintf(stderr, "ERROR: cannot have product columns of %d.\n",
            product_columns);

        exit(program_failure_code);

    } /* if (product_columns < minimum_matrix_size) */

    if (product_betweens < minimum_matrix_size) {

        fprintf(stderr,
            "ERROR: cannot have product betweens of %d.\n",
            product_betweens);

        exit(program_failure_code);

    } /* if (product_betweens < minimum_matrix_size) */

    scanf("%d", &multiply_method);

    if ((multiply_method < minimum_multiply_method) ||
        (multiply_method > maximum_multiply_method)) {

        fprintf(stderr,
            "ERROR: cannot have multiply method %d.\n",
            multiply_method);

        exit(program_failure_code);

    } /* if ((multiply_method < ...) || ...) */

    if (multiply_method == by_tiling) {

        scanf("%d %d %d",
              &rows_per_tile, &columns_per_tile, &betweens_per_tile);

        if (rows_per_tile < minimum_matrix_size) {
            fprintf(stderr,
                "ERROR: cannot have rows per tile of %d.\n",
                rows_per_tile);
            exit(program_failure_code);
        } /* if (rows_per_tile < minimum_matrix_size) */

        if (rows_per_tile > product_rows) {
            fprintf(stderr,
                "ERROR: cannot have rows per tile of %d",
                rows_per_tile);
            fprintf(stderr, " if product rows are %d.\n",
                product_rows);
            exit(program_failure_code);
        } /* if (rows_per_tile > product_rows) */

        if (columns_per_tile < minimum_matrix_size) {
            fprintf(stderr,
                "ERROR: cannot have columns per tile of %d.\n",
                columns_per_tile);
            exit(program_failure_code);
        } /* if (columns_per_tile < minimum_matrix_size) */

        if (columns_per_tile > product_columns) {
            fprintf(stderr,
                "ERROR: cannot have columns per tile of %d",
                columns_per_tile);
            fprintf(stderr, " if product columns are %d.\n",
                product_columns);
            exit(program_failure_code);
        } /* if (columns_per_tile > product_columns) */

        if (betweens_per_tile < minimum_matrix_size) {
            fprintf(stderr,
                "ERROR: cannot have betweens per tile of %d.\n",
                betweens_per_tile);
            exit(program_failure_code);
        } /* if (betweens_per_tile < minimum_matrix_size) */

        if (betweens_per_tile > product_betweens) {
            fprintf(stderr,
                "ERROR: cannot have betweens per tile of %d",
                betweens_per_tile);
            fprintf(stderr, " if product betweens are %d.\n",
                product_betweens);
            exit(program_failure_code);
        } /* if (betweens_per_tile > product_betweens) */

    } /* if (multiply_method == by_tiling) */

} /* input_sizes */

float** matrix_allocate (int matrix_rows, int matrix_columns)
{ /* matrix_allocate */
    float** matrix = (float**)NULL;
    int     row;

    matrix = (float**)malloc(matrix_rows * sizeof(float*));

    if (matrix == (float**)NULL) {

        fprintf(stderr,
            "ERROR: cannot allocate matrix of size %d rows.\n",
            matrix_rows);

        exit(program_failure_code);

    } /* if (matrix == (float**)NULL) */

    matrix[0] = (float* )malloc(matrix_rows * matrix_columns *
                                 sizeof(float));

    if (matrix[0] == (float*)NULL) {

        fprintf(stderr, "ERROR: cannot allocate matrix of size ");
        fprintf(stderr, "%d rows by %d columns.\n",
            matrix_rows, matrix_columns);

        exit(program_failure_code);

    } /* if (matrix[0] == (float*)NULL) */

    for (row = 0; row < matrix_rows; row++) {

        matrix[row] = &(matrix[0][row * matrix_columns]);

    } /* for row */

    return matrix;

} /* matrix_allocate */

float** matrix_deallocate (float** matrix,
                           int matrix_rows, int matrix_columns)
{ /* matrix_deallocate */

    free(matrix[0]); matrix[0] = (float* )NULL;

    free(matrix);    matrix    = (float**)NULL;

    return (float**)NULL;

} /* matrix_deallocate */

void matrix_set_randomly (float** matrix,
                          int number_of_rows, int number_of_columns)
{ /* matrix_set_randomly */
    int row, column;

    for (row = 0; row < number_of_rows; row++) {

        for (column = 0; column < number_of_columns; column++) {

            matrix[row][column] = 1.0 * random() / ((long long)RAND_MAX + 1);

        } /* for column */

    } /* for row */

} /* matrix_set_randomly */

void matrix_matrix_mult_by_init (
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens)
{ /* matrix_matrix_mult_by_init */
    double start_time, stop_time, multiply_time;
    int r, c, b;
    int total_number_of_elements, total_number_of_bytes;

    start_time = second();

    for (r = 0; r < product_rows; r++) {

        for (c = 0; c < product_columns; c++) {

            product[r][c] = 0.0;

            for (b = 0; b < product_betweens; b++) {

                product[r][c] += matrix1[r][b] * matrix2[b][c];

            } /* for b */

        } /* for c */

    } /* for r */

    stop_time = second();

    multiply_time = stop_time - start_time;

    total_number_of_elements =
        product_rows     * product_columns  +
        product_rows     * product_betweens +
        product_betweens * product_columns;
    total_number_of_bytes =
        total_number_of_elements * sizeof(float);

    printf("init:      %5d x %5d x %5d (%12d bytes): %g sec\n",
        product_rows, product_columns, product_betweens,
        total_number_of_bytes, multiply_time);

    printf("\n");

} /* matrix_matrix_mult_by_init */

void matrix_matrix_mult_by_tiling (
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens,
         int rows_per_tile, int columns_per_tile, int betweens_per_tile)
{ /* matrix_matrix_mult_by_tiling */
    double start_time, stop_time, multiply_time;
    int row_start, row_end;
    int column_start, column_end;
    int between_start, between_end;
    int r, c, b;
    int total_number_of_elements, total_number_of_bytes;
    int tile_number_of_elements,  tile_number_of_bytes;

    start_time = second();

    for (row_start = 0;
         row_start < product_rows;
         row_start += rows_per_tile) {

        row_end = row_start + rows_per_tile - 1;

        if (row_end >= product_rows) {
            row_end = product_rows - 1;
        } /* if (row_end >= product_rows) */

        for (column_start = 0;
             column_start < product_columns;
             column_start += columns_per_tile) {

            column_end = column_start + columns_per_tile - 1;

            if (column_end >= product_columns) {
                column_end = product_columns - 1;
            } /* if (column_end >= product_columns) */

            for (between_start = 0;
                 between_start < product_betweens;
                 between_start += betweens_per_tile) {

                between_end = between_start + betweens_per_tile - 1;

                if (between_end >= product_betweens) {
                    between_end = product_betweens - 1;
                } /* if (between_end >= product_betweens) */

                matrix_matrix_multiply_tile(
                    product, matrix1, matrix2,
                    product_rows, product_columns,
                    product_betweens,
                    row_start,     row_end,
                    column_start,  column_end,
                    between_start, between_end);

            } /* for between_start */

        } /* for column_start */

    } /* for row_start */

    stop_time  = second();

    multiply_time = stop_time - start_time;

    total_number_of_elements =
        product_rows    * product_columns +
        product_rows    * product_betweens +
        product_betweens * product_columns;
    total_number_of_bytes =
        total_number_of_elements * sizeof(float);

    tile_number_of_elements =
        rows_per_tile     * columns_per_tile  +
        rows_per_tile     * betweens_per_tile +
        betweens_per_tile * columns_per_tile;
    tile_number_of_bytes =
        tile_number_of_elements * sizeof(float);

    printf("tiling:    %5d x %5d x %5d (%12d bytes): %g\n",
        product_rows, product_columns, product_betweens,
        total_number_of_bytes, multiply_time);

    printf("  tile:    %5d x %5d x %5d (%12d bytes): %g\n",
        rows_per_tile, columns_per_tile, betweens_per_tile,
        tile_number_of_bytes, multiply_time);

    printf("\n");

} /* matrix_matrix_mult_by_tiling */

void matrix_matrix_multiply_tile (
         float** product, float** matrix1, float** matrix2,
         int product_rows, int product_columns, int product_betweens,
         int row_start,     int row_end,
         int column_start,  int column_end,
         int between_start, int between_end)
{ /* matrix_matrix_multiply_tile */

    int r, c, b;

    for (r = row_start; r <= row_end; r++) {

        for (c = column_start; c <= column_end; c++) {

            if (between_start == 0) product[r][c] = 0.0;

            for (b = between_start; b <= between_end; b++) {

                product[r][c] += matrix1[r][b] * matrix2[b][c];

            } /* for b */

        } /* for c */

    } /* for r */

} /* matrix_matrix_multiply_tile */

