      PROGRAM matrix_matrix_multiply_test

          IMPLICIT NONE

          INTEGER,PARAMETER :: minimum_matrix_size     = 1
          INTEGER,PARAMETER :: minimum_multiply_method = 1
          INTEGER,PARAMETER :: by_init      = minimum_multiply_method
          INTEGER,PARAMETER :: by_tiling    = by_init    + 1
          INTEGER,PARAMETER :: by_intrinsic = by_tiling  + 1
          INTEGER,PARAMETER :: maximum_multiply_method = by_intrinsic
          INTEGER,PARAMETER :: memory_success = 0

          REAL,DIMENSION(:,:),ALLOCATABLE :: product
          REAL,DIMENSION(:,:),ALLOCATABLE :: matrix1, matrix2

          INTEGER :: bits_per_real, bytes_per_real
          INTEGER :: multiply_method
          INTEGER :: product_rows, product_columns, product_betweens
          INTEGER :: rows_per_tile, columns_per_tile, betweens_per_tile
          INTEGER :: memory_status

!........ Determine the number of bits and bytes per REAL.

          CALL get_bits_and_bytes_per_real(  &
     &              bits_per_real, bytes_per_real)

!........ Input the sizes of the matrices, the type of multiplication,
!........ and the sizes of the tiles (if tiling).

          CALL input_sizes(                  &
     &             product_rows,             &
     &             product_columns,          &
     &             product_betweens,         &
     &             multiply_method,          &
     &             rows_per_tile,            &
     &             columns_per_tile,         &
     &             betweens_per_tile,        &
     &             minimum_matrix_size,      &
     &             minimum_multiply_method,  &
     &             maximum_multiply_method,  &
     &             by_tiling)

!........ Allocate the matrices.

          ALLOCATE(product(product_rows,product_columns),             &
     &             STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*) "ERROR: cannot allocate product of size ",  &
     &            product_rows, " rows by ",                          &
     &            product_columns, " columns (",                      &
     &            product_rows * product_columns *                    &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

          ALLOCATE(matrix1(product_rows,product_betweens),            &
     &             STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*) "ERROR: cannot allocate matrix1 of size ",  &
     &            product_rows, " rows by ",                          &
     &            product_betweens, " columns (",                     &
     &            product_rows * product_betweens *                   &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

          ALLOCATE(matrix2(product_betweens,product_columns),         &
     &             STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*) "ERROR: cannot allocate matrix1 of size ",  &
     &            product_betweens, " rows by ",                      &
     &            product_columns, " columns (",                      &
     &            product_betweens * product_columns *                &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

!........ Set the multiplier and multiplicand matrices at random.

          CALL matrix_set_randomly(matrix1,  &
     &             product_rows, product_betweens)

          CALL matrix_set_randomly(matrix2,  &
     &             product_betweens, product_columns)

          IF (multiply_method == by_init) THEN

!............ Perform the matrix-matrix multiply via initialization.

              CALL matrix_matrix_mult_by_init(           &
     &                 product, matrix1, matrix2,        &
     &                 product_rows, product_columns,    &
     &                 product_betweens,                 &
     &                 bytes_per_real)

          ELSE IF (multiply_method == by_intrinsic) THEN

!............ Perform the matrix-matrix multiply via the Fortran90
!............ intrinsic routine MATMUL.

              CALL matrix_matrix_mult_by_intrinsic(      &
     &                 product, matrix1, matrix2,        &
     &                 product_rows, product_columns,    &
     &                 product_betweens,                 &
     &                 bytes_per_real)

          ELSE IF (multiply_method == by_tiling) THEN

!............ Perform the matrix-matrix multiply via tiling.

              CALL matrix_matrix_mult_by_tiling(         &
     &                 product, matrix1, matrix2,        &
     &                 product_rows, product_columns,    &
     &                 product_betweens,                 &
     &                 rows_per_tile, columns_per_tile,  &
     &                 betweens_per_tile,                &
     &                 bytes_per_real)

          END IF !! (multiply_method == by_tiling)

!........ Deallocate the matrices.

          DEALLOCATE(product, STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*)  &
     &            "ERROR: cannot deallocate product of size ",  &
     &            product_rows, " rows by ",                    &
     &            product_columns, " columns (",                &
     &            product_rows * product_columns *              &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

          DEALLOCATE(matrix1, STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*)  &
     &            "ERROR: cannot deallocate matrix1 of size ",  &
     &            product_rows, " rows by ",                    &
     &            product_betweens, " columns (",               &
     &            product_rows * product_betweens *             &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

          DEALLOCATE(matrix2, STAT=memory_status)
          IF (memory_status /= memory_success) THEN
              WRITE (0,*)  &
     &            "ERROR: cannot deallocate matrix1 of size ",  &
     &            product_betweens, " rows by ",                &
     &            product_columns, " columns (",                &
     &            product_betweens * product_columns *          &
     &            bytes_per_real, " bytes)."
              STOP
          END IF !! (memory_status /= memory_success)

      END PROGRAM matrix_matrix_multiply_test

      SUBROUTINE get_bits_and_bytes_per_real (  &
     &               bits_per_real, bytes_per_real)

          IMPLICIT NONE

          INTEGER,INTENT(OUT) :: bits_per_real, bytes_per_real

          REAL,   PARAMETER :: round_increment = 0.499999
          INTEGER,PARAMETER :: bits_per_byte = 8
          INTEGER,PARAMETER :: number_of_sign_bits = 1

          REAL :: real_dummy

          bits_per_real =                                  &
     &        number_of_sign_bits + DIGITS(real_dummy) +   &
     &        NINT(LOG10(1.0 * MAXEXPONENT(real_dummy)) /  &
     &             LOG10(2.0) +                            &
     &             round_increment)
          bytes_per_real = bits_per_real / bits_per_byte

      END SUBROUTINE get_bits_and_bytes_per_real

      SUBROUTINE input_sizes (                 &
     &               product_rows,             &
     &               product_columns,          &
     &               product_betweens,         &
     &               multiply_method,          &
     &               rows_per_tile,            &
     &               columns_per_tile,         &
     &               betweens_per_tile,        &
     &               minimum_matrix_size,      &
     &               minimum_multiply_method,  &
     &               maximum_multiply_method,  &
     &               by_tiling)

          IMPLICIT NONE

          INTEGER,INTENT(OUT) :: product_rows, product_columns,  &
     &                           product_betweens
          INTEGER,INTENT(OUT) :: multiply_method
          INTEGER,INTENT(OUT) :: rows_per_tile, columns_per_tile,  &
     &                           betweens_per_tile
          INTEGER,INTENT(IN)  :: minimum_matrix_size
          INTEGER,INTENT(IN)  :: minimum_multiply_method
          INTEGER,INTENT(IN)  :: maximum_multiply_method
          INTEGER,INTENT(IN)  :: by_tiling

          READ *, product_rows, product_columns, product_betweens

          IF (product_rows < minimum_matrix_size) THEN
              WRITE (0,*)                                 &
     &            "ERROR: cannot have product rows of ",  &
     &            product_rows, "."
              STOP
          END IF !! (product_rows < minimum_matrix_size)

          IF (product_columns < minimum_matrix_size) THEN
              WRITE (0,*)                                 &
     &            "ERROR: cannot have product columns of ",  &
     &            product_columns, "."
              STOP
          END IF !! (product_columns < minimum_matrix_size)

          IF (product_betweens < minimum_matrix_size) THEN
              WRITE (0,*)                                 &
     &            "ERROR: cannot have product betweens of ",  &
     &            product_betweens, "."
              STOP
          END IF !! (product_betweens < minimum_matrix_size)

          READ *, multiply_method

          IF ((multiply_method < minimum_multiply_method) .OR.  &
     &        (multiply_method > maximum_multiply_method)) THEN
              WRITE (0,*)                                 &
     &            "ERROR: cannot have multiply method ",  &
     &            multiply_method, "."
              STOP
          END IF !! ((multiply_method < ...) .OR. ...)

          IF (multiply_method == by_tiling) THEN

              READ *, rows_per_tile, columns_per_tile,  &
     &                betweens_per_tile

              IF (rows_per_tile < minimum_matrix_size) THEN
                  WRITE (0,*)                                  &
     &                "ERROR: cannot have rows per tile of ",  &
     &                rows_per_tile, "."
                  STOP
              END IF !! (rows_per_tile < minimum_matrix_size)

              IF (rows_per_tile > product_rows) THEN
                  WRITE (0,*)                                  &
     &                "ERROR: cannot have rows per tile of ",  &
     &                rows_per_tile, " if product rows are ",  &
     &                product_rows, "."
                  STOP
              END IF !! (rows_per_tile > product_rows)

              IF (columns_per_tile < minimum_matrix_size) THEN
                  WRITE (0,*)                                     &
     &                "ERROR: cannot have columns per tile of ",  &
     &                columns_per_tile, "."
                  STOP
              END IF !! (columns_per_tile < minimum_matrix_size)

              IF (columns_per_tile > product_columns) THEN
                  WRITE (0,*)                                        &
     &                "ERROR: cannot have columns per tile of ",     &
     &                columns_per_tile, " if product columns are ",  &
     &                product_columns, "."
                  STOP
              END IF !! (columns_per_tile > product_columns)

              IF (betweens_per_tile < minimum_matrix_size) THEN
                  WRITE (0,*)                                      &
     &                "ERROR: cannot have betweens per tile of ",  &
     &                betweens_per_tile, "."
                  STOP
              END IF !! (betweens_per_tile < minimum_matrix_size)

              IF (betweens_per_tile > product_betweens) THEN
                  WRITE (0,*)                                         &
     &                "ERROR: cannot have betweens per tile of ",     &
     &                betweens_per_tile, " if product betweens are ", &
     &                product_betweens, "."
                  STOP
              END IF !! (betweens_per_tile > product_betweens)

          END IF !! (multiply_method == by_tiling)

      END SUBROUTINE input_sizes

      SUBROUTINE matrix_set_randomly (matrix,  &
     &               number_of_rows, number_of_columns)

          IMPLICIT NONE

          INTEGER,INTENT(IN) :: number_of_rows, number_of_columns
          REAL,DIMENSION(number_of_rows,number_of_columns),  &
     &         INTENT(OUT) :: matrix

          CALL RANDOM_NUMBER(  &
     &             matrix(1:number_of_rows,1:number_of_columns))

      END SUBROUTINE matrix_set_randomly

      SUBROUTINE matrix_matrix_mult_by_init (        &
     &               product, matrix1, matrix2,      &
     &               product_rows, product_columns,  &
     &               product_betweens,               &
     &               bytes_per_real)

          IMPLICIT NONE

          INTEGER,INTENT(IN) :: product_rows, product_columns,  &
     &                          product_betweens
          REAL,DIMENSION(product_rows,product_columns),         &
     &         INTENT(OUT) :: product
          REAL,DIMENSION(product_rows,product_betweens),        &
     &         INTENT(IN)  :: matrix1
          REAL,DIMENSION(product_betweens,product_columns),     &
     &         INTENT(IN)  :: matrix2
          INTEGER,INTENT(IN) :: bytes_per_real

          REAL    :: start_time, stop_time, multiply_time
          INTEGER :: r, c, b
          INTEGER :: total_number_of_elements, total_number_of_bytes

          REAL*8,EXTERNAL :: timings

          start_time = timings()

          DO c = 1, product_columns

              DO r = 1, product_rows

                  product(r,c) = 0.0

                  DO b = 1, product_betweens

                      product(r,c) =  &
     &                    product(r,c) + matrix1(r,b) * matrix2(b,c)

                  END DO !! b

              END DO !! r

          END DO !! c

          stop_time  = timings()

          multiply_time = stop_time - start_time

          total_number_of_elements =                &
              product_rows    * product_columns +   &
     &        product_rows    * product_betweens +  &
     &        product_betweens * product_columns
          total_number_of_bytes =  &
     &        total_number_of_elements * bytes_per_real

          PRINT "(A,I5,A,I5,A,I5,A,I12,A,F20.10,A)",     &
     &        "init:      ",                             &
     &        product_rows,    " x ",                    &
     &        product_columns, " x ",                    &
     &        product_betweens,                          &
     &        " (", total_number_of_bytes, " bytes): ",  &
     &        multiply_time, " sec"

          PRINT *, ""

      END SUBROUTINE matrix_matrix_mult_by_init

      SUBROUTINE matrix_matrix_mult_by_intrinsic (   &
     &               product, matrix1, matrix2,      &
     &               product_rows, product_columns,  &
     &               product_betweens,               &
     &               bytes_per_real)

          IMPLICIT NONE

          INTEGER,INTENT(IN) :: product_rows, product_columns,  &
     &                          product_betweens

          REAL,DIMENSION(product_rows,   product_columns),   &
     &         INTENT(OUT) :: product
          REAL,DIMENSION(product_rows,   product_betweens),  &
     &         INTENT(IN)  :: matrix1
          REAL,DIMENSION(product_betweens,product_columns),  &
     &         INTENT(IN)  :: matrix2
          INTEGER,INTENT(IN) :: bytes_per_real

          REAL    :: start_time, stop_time, multiply_time
          INTEGER :: total_number_of_elements, total_number_of_bytes

          REAL*8,EXTERNAL :: timings

          start_time = timings()

          product = MATMUL(matrix1, matrix2)

          stop_time  = timings()

          multiply_time = stop_time - start_time

          total_number_of_elements =                &
              product_rows    * product_columns +   &
     &        product_rows    * product_betweens +  &
     &        product_betweens * product_columns
          total_number_of_bytes =  &
     &        total_number_of_elements * bytes_per_real

          PRINT "(A,I5,A,I5,A,I5,A,I12,A,G25.10)",  &
     &        "intrinsic: ",                        &
     &        product_rows, " x ",                  &
     &        product_columns, " x ",               &
     &        product_betweens,                     &
     &        " (", total_number_of_bytes, "):", multiply_time

          PRINT *, ""

      END SUBROUTINE matrix_matrix_mult_by_intrinsic

      SUBROUTINE matrix_matrix_mult_by_tiling (        &
     &               product, matrix1, matrix2,        &
     &               product_rows, product_columns,    &
     &               product_betweens,                 &
     &               rows_per_tile, columns_per_tile,  &
     &               betweens_per_tile,                &
     &               bytes_per_real)

          IMPLICIT NONE

          INTEGER,INTENT(IN) :: product_rows, product_columns,    &
     &                          product_betweens
          REAL,DIMENSION(product_rows,   product_columns),        &
     &         INTENT(OUT) :: product
          REAL,DIMENSION(product_rows,   product_betweens),       &
     &         INTENT(IN)  :: matrix1
          REAL,DIMENSION(product_betweens,product_columns),       &
     &         INTENT(IN)  :: matrix2
          INTEGER,INTENT(IN) :: rows_per_tile, columns_per_tile,  &
     &                          betweens_per_tile
          INTEGER,INTENT(IN) :: bytes_per_real

          REAL    :: start_time, stop_time, multiply_time
          INTEGER :: row_start, row_end
          INTEGER :: column_start, column_end
          INTEGER :: between_start, between_end
          INTEGER :: r, c, b
          INTEGER :: total_number_of_elements, total_number_of_bytes
          INTEGER :: tile_number_of_elements,  tile_number_of_bytes

          REAL*8,EXTERNAL :: timings

          start_time = timings()

          DO column_start = 1, product_columns, columns_per_tile

              column_end = column_start + columns_per_tile - 1
              IF (column_end > product_columns) THEN
                  column_end = product_columns
              END IF !! (column_end > product_columns)

              DO row_start = 1, product_rows, rows_per_tile

                  row_end = row_start + rows_per_tile - 1
                  IF (row_end > product_rows) THEN
                      row_end = product_rows
                  END IF !! (row_end > product_rows)

                  DO between_start = 1, product_betweens,  &
     &                               betweens_per_tile

                      between_end =   &
     &                    between_start + betweens_per_tile - 1
                      IF (between_end > product_betweens) THEN
                          between_end = product_betweens
                      END IF !! (between_end > product_betweens)

                      CALL matrix_matrix_multiply_tile(        &
     &                         product, matrix1, matrix2,      &
     &                         product_rows, product_columns,  &
     &                         product_betweens,               &
     &                         row_start,     row_end,         &
     &                         column_start,  column_end,      &
     &                         between_start, between_end)

                  END DO !! between_start

              END DO !! row_start

          END DO !! column_start

          stop_time  = timings()

          multiply_time = stop_time - start_time

          total_number_of_elements =                &
              product_rows    * product_columns +   &
     &        product_rows    * product_betweens +  &
     &        product_betweens * product_columns
          total_number_of_bytes =  &
     &        total_number_of_elements * bytes_per_real

          tile_number_of_elements =                    &
              rows_per_tile     * columns_per_tile  +  &
     &        rows_per_tile     * betweens_per_tile +  &
     &        betweens_per_tile * columns_per_tile
          tile_number_of_bytes =  &
     &        tile_number_of_elements * bytes_per_real

          PRINT "(A,I5,A,I5,A,I5,A,I12,A,F20.10)",            &
     &        "tiling:    ",                                  &
     &        product_rows,     "x",                          &
     &        product_columns,  "x",                          &
     &        product_betweens,  &
     &        " (", total_number_of_bytes, " bytes): ", multiply_time

          PRINT "(A,I5,A,I5,A,I5,A,I12,A,F20.10)",  &
     &        "  tile: ",                           &
     &        rows_per_tile,    "x",                &
     &        columns_per_tile, "x",                &
     &        betweens_per_tile,                    &
     &        " (", tile_number_of_bytes,  " bytes): ", multiply_time

          PRINT *, ""

      END SUBROUTINE matrix_matrix_mult_by_tiling

      SUBROUTINE matrix_matrix_multiply_tile (       &
     &               product, matrix1, matrix2,      &
     &               product_rows, product_columns,  &
     &               product_betweens,               &
     &               row_start,     row_end,         &
     &               column_start,  column_end,      &
     &               between_start, between_end)

          IMPLICIT NONE

          INTEGER,INTENT(IN) :: product_rows, product_columns,  &
     &                          product_betweens

          REAL,DIMENSION(product_rows,   product_columns),   &
     &         INTENT(OUT) :: product
          REAL,DIMENSION(product_rows,   product_betweens),  &
     &         INTENT(IN)  :: matrix1
          REAL,DIMENSION(product_betweens,product_columns),  &
     &         INTENT(IN)  :: matrix2
          INTEGER,INTENT(IN) :: row_start,     row_end
          INTEGER,INTENT(IN) :: column_start,  column_end
          INTEGER,INTENT(IN) :: between_start, between_end

          INTEGER :: r, c, b

          DO c = column_start, column_end

              DO r = row_start, row_end

                  IF (between_start == 1) product(r,c) = 0.0

                  DO b = between_start, between_end

                      product(r,c) =  &
     &                    product(r,c) + matrix1(r,b) * matrix2(b,c)

                  END DO !! b = between_start, between_end

              END DO !! r = row_start, row_end

          END DO !! c = column_start, column_end

      END SUBROUTINE matrix_matrix_multiply_tile

